import logging
import random

import pandas as pd

from time_blender.util import fresh_id

import pymc3 as pm


class Event:

    def __init__(self, name=None, parallel_events=None, push_down=False):
        self.name = name

        self._indexed_generated_values = {} # time to value
        self._generated_values = [] # sequence of values
        self._last_pos = -1

        # determine which attributes are causal events
        self._causal_parameters = []
        for k, v in self.__dict__.items():
            if isinstance(v, Event):
                self._causal_parameters.append(v)

        if parallel_events is not None:
            self.parallel_events = []
            self.parallel_to(parallel_events)
        else:
            self.parallel_events = None

        self.push_down = push_down

        self._execution_locked = False

    def execute(self, t):
        """
        Executes the event and generates an output for the present moment.

        :param t: The time in which the event takes place.
        :return: The scalar value of the event in the specified moment.
        """

        if not self._execution_locked:
            self._execution_locked = True

            # update parallel events
            if self.parallel_events is not None:
                for e in self.parallel_events:
                    e.execute(t)

            # process this event
            if t not in self._indexed_generated_values:
                self._last_pos += 1
                # save the value for future reference and to avoid recomputing
                v = self._execute(t, self._last_pos)
                self._indexed_generated_values[t] = v
                self._generated_values.append(v)

            res = self._indexed_generated_values[t]

            # Result might be used by underlying events as well. So it must be pushed down.
            self._push_down(t, res)

            self._execution_locked = False

            return res

        else:
            return None

    def _execute(self, t, i):
        raise NotImplementedError("Must be implemented by concrete subclasses.")

    def _push_down(self, t, parent_value):
        """
        Given a value executed at the specified moment by a parent event, pushes it down to its children
        events. That is to say, provides a downward path for executed values, opposite to the the regular
        information flow from children to parent.

        :param t: The execution moment.
        :param parent_value: The value executed by a parent.
        :return: None
        """
        self._capture_push_down_value(t, parent_value)

        if self.push_down:
            for e in self._causal_parameters:
                e._push_down(t, parent_value)

    def _capture_push_down_value(self, t, parent_value):
        """
        Receives a value executed by a parent at the specified time and, if needed, stores this value locally.
        By default, actually nothing is stored, concrete subclasses must overload this method to do so.
        For example, events that are stateful and supposed to track some current value might store values
        produced by parents.

        :param t:
        :param parent_value:
        :return:
        """
        pass

    def _value_or_execute_if_event(self, x, t):
        if isinstance(x, Event):
            return x.execute(t)
        else:
            return x

    def parallel_to(self, events):
        if self.parallel_events is None:
            self.parallel_events = []

        if isinstance(events, list):
            for pe in events:
                if isinstance(pe, Event):
                    self.parallel_events.append(pe)
                    self._causal_parameters.append(pe)
        elif isinstance(events, Event):
            self.parallel_events.append(events)
            self._causal_parameters.append(events)
        else:
            raise ValueError("Either a list of events or an event must be specified.")

    def value_at_pos(self, i):
        return self._generated_values[i]

    def value_at(self, t):
        return self._indexed_generated_values[t]

    def _default_name_if_none(self, name=None):
        if name is None:
            return f"{type(self).__name__}_{str(fresh_id())}"
        else:
            return name

    def reset(self):
        """
        Cleans all caches in order to allow the reuse of the event in a new generation process.
        :return: None
        """
        self._indexed_generated_values = {}
        self._generated_values = []
        self._last_pos = -1

        # clear causes too, if the object is not locked
        if not self._execution_locked:
            self._execution_locked = True

            for e in self._causal_parameters:
                e.reset()

            self._execution_locked = False

    def _pymc3_model_variables(self, model, observations=None):
        found_random_event = False
        for event in self._causal_parameters:
            if not found_random_event:
                event._pymc3_model_variables(model, observations)
                if isinstance(event, RandomEvent):
                    found_random_event = True
            else:
                # There can be only one root RandomEvent because observations must be related to exactly one
                # event in the causal graph.
                raise ValueError("There can be only one root random event (all other random events must be causally \
                                  linked to it)")

    def _pymc3_push_distributions_down(self, trace):
        for event in self._causal_parameters:
            event._pymc3_push_distributions_down(trace)

    def generalize_from_observations(self, observations):
        """
        Given various observations, learn the model parameters that best fit them.

        :param observations: A sequence observations. E.g., [1, 0, 1, 0, 1, 1].
        :return:
        """
        raise NotImplementedError("Only random events can be subject to learning.")

    def is_root_cause(self):
        """
        Checks whether this event depend on any other or not (i.e., it is a root cause).
        :return: True if no dependecies exist; False otherwise.
        """
        return len(self._causal_parameters) == 0

    def __add__(self, other):
            return LambdaEvent(lambda t, i, mem: self.execute(t) + other.execute(t), sub_events=[other])

    def __sub__(self, other):
        return LambdaEvent(lambda t, i, mem: self.execute(t) - other.execute(t), sub_events=[other])

    def __mul__(self, other):
        return LambdaEvent(lambda t, i, mem: self.execute(t) * other.execute(t), sub_events=[other])

    def __truediv__(self, other):
        return LambdaEvent(lambda t, i, mem: self.execute(t) / other.execute(t), sub_events=[other])


class RandomEvent(Event):

    def __init__(self, pymc3_distribution_func, name=None, parallel_events=None, push_down=False):
        self.pymc3_distribution_func = pymc3_distribution_func
        self.distribution_sample = []
        super().__init__(name, parallel_events, push_down)

    def _pymc3_model_variables(self, model, observations=None):
        return self.pymc3_distribution_func(model, observations)

    @staticmethod
    def _pymc3_model_variables_if_distribution(model, x, observations=None):
        if isinstance(x, RandomEvent):
            return x._pymc3_model_variables(model, observations)
        elif isinstance(x, Event):
            raise ValueError("Only random events or scalars are allowed.")
        else:
            return x

    def _pymc3_push_distributions_down(self, trace):
        try:
            self.distribution_sample = trace[self.name]
        except KeyError as err:
            logging.debug(err)

        super()._pymc3_push_distributions_down(trace)

    def generalize_from_observations(self, observations):
        """
        Given various observations, learn the model parameters that best fit them.

        :param observations: A sequence observations. E.g., [1, 0, 1, 0, 1, 1].
        :return:
        """
        model = pm.Model()
        with model:
            self._pymc3_model_variables(model, observations)

            # Calculate a posteriori distribution from observations
            step = pm.Metropolis()
            self.trace = pm.sample(10000, tune=5000, step=step) # TODO customize parameters and algorithm

            # propagate the new distributions down
            self._pymc3_push_distributions_down(self.trace)

    def sample_from_learned_distribution(self):
        """
        Returns a sample based on the (a posteriori) learned distribution.
        :return: a scalar value.
        """
        if len(self.distribution_sample) == 0:
            raise ValueError("Cannot pick from empty distribution sample.")

        return random.choice(self.distribution_sample)

    def sample_from_definition(self, t):
        raise NotImplementedError("Must be implemented by concrete subclasses.")

    def has_learned(self):
        """
        Checks whether this event has been subject to learning from data.
        :return: True if learning took place; False otherwise.
        """
        return len(self.distribution_sample) > 0

    def _should_use_learned_sample(self):
        return self.is_root_cause() and self.has_learned()

    def _execute(self, t, i):
        if self._should_use_learned_sample():
            return self.sample_from_learned_distribution()
        else:
            return self.sample_from_definition(t)


class LambdaEvent(Event):

    def __init__(self, func, sub_events=[], name=None, parallel_events=None, push_down=False):
        super().__init__(name, parallel_events, push_down)
        self.func = func
        self.memory = {}

        for se in sub_events:
            if isinstance(se, Event):
                self._causal_parameters.append(se)

    def _execute(self, t, i):
        return self.func(t, i, self.memory)

    def reset(self):
        self.memory = {}
        super().reset()


class Generator:

    def __init__(self, events):
        """
        Creates a new time series generator using the specified events. The names of the events will be used
        later to name the columns of generated DataFrames.

        :param events: Either a list of events or a dict of events. In the former case, the name of each
        event is retrieved from the event itself. In the latter case, the user can specifiy new names.
        """
        if isinstance(events, dict):
            self.named_events = events
        else:
            self.named_events = {}

            # make iterable
            if not isinstance(events, list):
                events = [events]

            for i, event in enumerate(events):
                if event.name is not None:
                    self.named_events[event.name] = event
                else:
                    self.named_events['Event ' + str(i)] = event

    def generate(self, start_date, end_date, n=1, between_time=None, freq='D', filter_func=lambda df: df):
        """
        Generates time series from the model assigned to the present generator.

        :param start_date: The first date of the series to generate.
        :param end_date: The last date of the series to generate.
        :param n: The number of series to generate.
        :param freq: The frequency to be used, in terms of Pandas frequency strings.
        :param filter_func: A filter function to apply to the generated data before returning.

        :return: A list of generated time series.
        """
        generated_data = []
        for i in range(0, n):
            values = {}
            dates = pd.date_range(start_date, end_date, freq=freq)
            for name in self.named_events:
                self.named_events[name].reset()  # clears the cache

            for t in dates:
                for name in self.named_events:
                    if name not in values:
                        values[name] = []

                    values[name].append(self.named_events[name].execute(t))

            df = pd.DataFrame(values, index=dates)
            if between_time is not None:
                df = df.between_time(between_time[0], between_time[1])

            df = filter_func(df)
            generated_data.append(df)

        return generated_data


def generate(model, start_date, end_date, n=1, freq='D', filter_func=lambda df: df):
    """
    A convenience method to generate time series from the specified model using the default generator.

    :param model: The model from which the data is to be generated.
    :param start_date: The first date of the series to generate.
    :param end_date: The last date of the series to generate.
    :param n: The number of series to generate.
    :param freq: The frequency to be used, in terms of Pandas frequency strings.
    :param filter_func: A filter function to apply to the generated data before returning.

    :return: A list of generated time series.
    """
    data = Generator(model).generate(start_date,
                                     end_date,
                                     n=n,
                                     freq=freq,
                                     filter_func=filter_func)

    return data


def save_to_csv(data, common_name, **kwargs):
    for i, df in enumerate(data):
        name = common_name + '_' + i
        df.to_csv(name, kwargs)



from time_blender.core import Event
import numpy as np


class IdentityEvent(Event):
    def __init__(self, event, name=None, parallel_events=None, push_down=False):
        super().__init__(name, parallel_events, push_down)
        self.event = event

    def _execute(self, t, i):
        return self.event.execute(t)


class ClipEvent(Event):
    def __init__(self, event, max_value=None, min_value=None, name=None, parallel_events=None, push_down=False):
        super().__init__(name, parallel_events, push_down)
        self.event = event
        self.max_value = max_value
        self.min_value = min_value

    def _execute(self, t, i):
        v = self.event.execute(t)

        if self.max_value is not None and v > self.max_value:
            v = self.max_value
        elif self.min_value is not None and v < self.min_value:
            v = self.min_value

        return v

class ClockEvent(Event):

    def __init__(self, as_ticks=True, name=None, parallel_events=None, push_down=False):
        super().__init__(name, parallel_events, push_down)
        self.as_ticks = as_ticks

    def _execute(self, t, i):
        if self.as_ticks:
            return i
        else:
            return t


class WaveEvent(Event):

    def __init__(self, period, amplitude, pos=0.0, name=None, parallel_events=None, push_down=False):
        super().__init__(name, parallel_events, push_down)
        self.period = period
        self.amplitude = amplitude

        self.pos = pos
        self.ini_pos = pos

    def _execute(self, t, i):
        step = 2*np.pi / self._value_or_execute_if_event(self.period, t)
        self.pos += step
        self.pos = self.pos % (2*np.pi)
        return self._value_or_execute_if_event(self.amplitude, t) * np.sin(self.pos)

    def reset(self):
        self.pos = self.ini_pos
        super().reset()


class ConstantEvent(Event):
    def __init__(self, constant, name=None, parallel_events=None, push_down=False):
        super().__init__(name, parallel_events, push_down)
        self.constant = constant

    def _execute(self, t, i):
        return self._value_or_execute_if_event(self.constant, t)


class WalkEvent(Event):

    def __init__(self, pos, step, name=None, parallel_events=None, push_down=False):
        super().__init__(name, parallel_events, push_down)
        self.step = step

        self.ini_pos = pos
        self.pos = pos

    def _execute(self, t, i):
        self.pos = self.pos + self._value_or_execute_if_event(self.step, t)
        return self.pos

    def _capture_push_down_value(self, t, parent_value):
        self.pos = parent_value

    def reset(self):
        self.pos = self.ini_pos
        super().reset()



from time_blender.core import RandomEvent
import numpy as np
import pymc3 as pm


class UniformEvent(RandomEvent):

    def __init__(self, low=-1.0, high=1.0, name=None, parallel_events=None, push_down=False):
        name = self._default_name_if_none(name)
        self.low = low
        self.high = high

        super().__init__(lambda model, obs: pm.Uniform(self.name,
                                                       lower=self._pymc3_model_variables_if_distribution(model, low),
                                                       upper=self._pymc3_model_variables_if_distribution(model, high),
                                                       observed=obs),
                         name, parallel_events, push_down)

    def sample_from_definition(self, t):
        return np.random.uniform(self._value_or_execute_if_event(self.low, t),
                                 self._value_or_execute_if_event(self.high, t))


class NormalEvent(RandomEvent):

    def __init__(self, mean, std, name=None, parallel_events=None, push_down=False):
        name = self._default_name_if_none(name)
        self.mean = mean
        self.std = std

        super().__init__(lambda model, obs: pm.Normal(self.name,
                                                      mu=self._pymc3_model_variables_if_distribution(model, mean),
                                                      sd=self._pymc3_model_variables_if_distribution(model, std),
                                                      observed=obs),
                         name, parallel_events, push_down)

    def sample_from_definition(self, t):
        return np.random.normal(self._value_or_execute_if_event(self.mean, t),
                                self._value_or_execute_if_event(self.std, t))


class PoissonEvent(RandomEvent):

    def __init__(self, lamb, name=None, parallel_events=None, push_down=False):
        name = self._default_name_if_none(name)
        self.lamb = lamb

        super().__init__(lambda model, obs: pm.Poisson(self.name,
                                                       mu=self._pymc3_model_variables_if_distribution(model, lamb),
                                                       observed=obs),
                         name, parallel_events, push_down)

    def sample_from_definition(self, t):
        return np.random.poisson(self._value_or_execute_if_event(self.lamb, t))


class Resistance(RandomEvent):

    def __init__(self, event, resistance_value_begin, resistance_value_end,
                 resistance_probability, resistance_strength_event, direction, name=None, parallel_events=None,
                 push_down=False):

        self.event = event
        self.resistance_value_begin = resistance_value_begin
        self.resistance_value_end = resistance_value_end
        self.resistance_probability = resistance_probability
        self.resistance_strength_event = resistance_strength_event
        self.direction = direction

        def aux(model, obs):
            raise NotImplementedError("Not supported.")

        super().__init__(aux,
                         name, parallel_events, push_down)

    def sample_from_definition(self, t):
        value = self.event.execute(t)

        # Decide whether to resist
        if self.direction == 'top' and \
           self.resistance_value_begin <= value < self.resistance_value_end and \
           self.resistance_probability > np.random.random():

                resist = True

        elif self.direction == 'bottom' and \
             self.resistance_value_end < value <= self.resistance_value_begin and \
             self.resistance_probability > np.random.random():

                resist = True
        else:
            resist = False

        # apply resistance, if pertinent
        if resist:
            resistance_factor = self.resistance_strength_event.execute(t)
            assert(0.0 <= resistance_factor <= 1.0)

            if self.direction == 'top':
                # pushes downward
                return value - abs(value) * resistance_factor

            else:
                # pushes upward
                return value + abs(value) * resistance_factor
        else:
            # no resistance applied
            return value


class TopResistance(Resistance):
    def __init__(self, event, resistance_value_begin, resistance_value_end, resistance_probability,
                 resistance_strength_event, name=None, parallel_events=None, push_down=True):
        super().__init__(event=event,
                         resistance_value_begin=resistance_value_begin,
                         resistance_value_end=resistance_value_end,
                         resistance_probability=resistance_probability,
                         resistance_strength_event=resistance_strength_event,
                         direction='top',
                         name=name, parallel_events=parallel_events, push_down=push_down)


class BottomResistance(Resistance):
    def __init__(self, event, resistance_value_begin, resistance_value_end, resistance_probability,
                 resistance_strength_event, name=None, parallel_events=None, push_down=True):
        super().__init__(event=event,
                         resistance_value_begin=resistance_value_begin,
                         resistance_value_end=resistance_value_end,
                         resistance_probability=resistance_probability,
                         resistance_strength_event=resistance_strength_event,
                         direction='bottom',
                         name=name, parallel_events=parallel_events, push_down=push_down)


# TODO bubble

# TODO mean-reverting models



########################################################################################################################
# Command-Line Interface utils
########################################################################################################################
import pandas as pd
from sigtools.wrappers import decorator, wrapper_decorator
from clize import run, parser, Parameter
import ast
import matplotlib.pyplot as plt

import time_blender
from time_blender.util import set_random_seed


@parser.value_converter
def a_list(x):
    if isinstance(x, list):
        return x
    else:
        return ast.literal_eval(x)


using_cli = False
cli_end_date = pd.Timestamp.now()
cli_begin_date = cli_end_date - pd.DateOffset(months=12)
cli_output_file = None #'out.csv'
cli_freq = 'D'
cli_plot = False


@wrapper_decorator
def cli_model(model_func, *args, n: int=1, begin=None, end=None, freq='D', output_file=None, plot=False,
              random_seed=None, **kwargs: Parameter.IGNORE):
    """

    :param model_func:
    :param args:
    :param n: How many series are to be generated with the specified parameters.
    :param begin: The first date to be generated. Format: YYYY-MM-DD
    :param end: The last date to be generated. Format: YYYY-MM-DD
    :param freq: The frequency of the generation. Refer to Pandas documentation for the proper string values.
    :param output_file: A prefix to be used as the base of files in which to save the series. If this is omitted,
                        the series is just returned in the console.
    :param plot: Whether the series should be plotted. If an output file is defined, the base name is also used
                 to save an image.
    :param kwargs:
    :return:
    """
    if using_cli:

        set_random_seed(random_seed)

        event = model_func(*args, **kwargs)

        if end is None:
            cli_end_date = pd.Timestamp.now()
        else:
            cli_end_date = pd.to_datetime(end)

        if begin is None:
            cli_begin_date = cli_end_date - pd.DateOffset(months=12)
        else:
            cli_begin_date = pd.to_datetime(begin)

        cli_freq = freq
        cli_output_file = output_file

        g = time_blender.core.Generator([event])
        print(cli_end_date, cli_end_date)
        data = g.generate(cli_begin_date, cli_end_date, n=n, freq=cli_freq)
        #print(data)
        results = None
        for i, d in enumerate(data):
            if cli_output_file is not None:
                if i == 0:
                    name = f'{cli_output_file}'
                else:
                    name = f'{i}_{cli_output_file}'

                d.to_csv(name + '.csv')
                plt.ioff()
                plt.clf()
                d.plot()
                plt.savefig(name + '.png')

            else:
                if results is None:
                    results = ""
                results += d.to_csv() # without a path, returns a string with the CSV

                if plot:
                    d.plot()


        # Shows all plots at once
        if cli_output_file is None:
            plt.show()

        return results

    else:
        return model_func(*args, **kwargs)




import random

import pymc3 as pm
import pandas as pd
import numpy as np

# A counter to be used when creating fresh names
import time_blender

def fresh_id():
    fresh_id.counter += 1
    return fresh_id.counter

fresh_id.counter = -1


def set_random_seed(seed):
    if seed is not None:
        random.seed(seed)
        np.random.seed(seed)
        # Add any other source of randomness here


def shift_weekend_and_holidays(day, direction='forward', holidays=[]):
    shift = 0
    if (day.weekday() == 5) or (day.weekday() == 6) or (day in holidays):
        if direction == 'forward':
            shift = 1
        else:
            shift = -1

        # recursive step
        return shift_weekend_and_holidays(day + pd.DateOffset(days=shift), direction, holidays)

    else:
        # recursion base
        return day


class PyMC3Utils:

    @staticmethod
    def multi_switch( vars, categorical_var):
        """
        Selects a variable in the specified set according to the categorical distribution provided.
        :param vars: a collection of PyMC3 random variables or scalars.
        :param categorical_var: a PyMC3 categorical variable.
        :return: a composite PyMC3 switch random variable.
        """
        vars = set(vars)

        def aux(vars, i=0):
            if len(vars) > 2:
                var = vars.pop()
                return pm.math.switch(categorical_var == i,
                                      var,
                                      aux(vars, i + 1))
            else:
                return vars.pop()

        return aux(vars)





# Standard models
import numpy as np
import pandas as pd

from time_blender.coordination_events import PastEvent
from time_blender.core import LambdaEvent
from time_blender.deterministic_events import ConstantEvent, WaveEvent, ClockEvent, WalkEvent, IdentityEvent, ClipEvent
from time_blender.random_events import NormalEvent

from clize import Parameter


from time_blender.util import shift_weekend_and_holidays
from time_blender.cli import cli_model, a_list


class SimpleModels:
    @staticmethod
    @cli_model
    def cycle(base:float=10.0, period:float=72, growth_rate:float=2):

        period_fluctuation = WalkEvent(period, NormalEvent(0, 1))
        amplitude_trend = ClockEvent() * ConstantEvent(base) * NormalEvent(3, 1)
        we = WaveEvent(period_fluctuation, amplitude_trend)

        capacity_trend = ClockEvent() * ConstantEvent(growth_rate*base) * NormalEvent(3, 0.1)

        return we + capacity_trend


class ClassicModels:

    @staticmethod
    @cli_model
    def ar(p: int, *, constant: float=0, error_mean: float=0, error_std: float=1, coefs_low: float=-1, coefs_high: float=1,
           coefs: a_list=None, error_event: Parameter.IGNORE=None):
        """
        Creates a new AR (autoreressive) model. The model's coeficients can either be generated automatically
        by providing coefs_low and coefs_high parameters, or be explicitly defined by providing a list in the
        coefs parameter.

        :param p: The order of the AR model (how far back should it look).
        :param constant: The model's constant term.
        :param error_mean: The mean of the normal error component.
        :param error_std: The standard deviation of the normal error component.
        :param coefs_low: If specified, defines the lower bound of the coeficients to be generated. If left None,
                          then the coefs parameter must be specified.
        :param coefs_high: If specified, defines the upper bound of the coeficients to be generated. If left None,
                          then the coefs parameter must be specified.
        :param coefs: A list or dict with numeric keys of the coeficients to be employed. Must have size p. The i-th
                      element (if list) or key (if dict) correspond to the i-th coeficient.
                      If this is specified, coefs_low and coefs_high are ignored.
        :param error_event: An error event. If specified, it is used instead of error_mean and error_std.
        :return: An AR model.
        """

        # check coeficients
        if (coefs is None) and (coefs_low is None) and (coefs_high is None):
            raise ValueError("Either coefs or coefs_loe, coefs_high must be specified.")

        # check error events
        if (coefs is error_mean) and (error_std is None) and (error_event is None):
            raise ValueError("Some error must be specified.")

        # Start with the model's constant.
        if error_event is not None:
            x = ConstantEvent(constant, parallel_events=[error_event])
        else:
            x = ConstantEvent(constant)

        past = []

        # Add the autoregressive terms
        for i in range(0, p):

            if coefs is not None:
                alpha = coefs[i]
            else:
                alpha = np.random.uniform(coefs_low, coefs_high)

            p = PastEvent(i + 1)
            past.append(p)

            if error_event is not None:
                error = PastEvent(i)
                error.refers_to(error_event)
            else:
                error = NormalEvent(error_mean, error_std)

            x = x + p * ConstantEvent(alpha) + error

        # connect past events to the series to which they refer to
        for p in past:
            p.refers_to(x)

        return x

    @staticmethod
    @cli_model
    def ma(q, *, series_mean: float=0, error_mean: float=0, error_std: float=1,
           coefs_low: float=-1, coefs_high: float=1, coefs: a_list=None, error_event: Parameter.IGNORE=None):
        """
        Creates a new MA (Moving Average) model. The model's coeficients can either be generated automatically
        by providing coefs_low and coefs_high parameters (default), or be explicitly defined by providing a list in the
        coefs parameter.

        :param q: The order of the MA model (how far back should it look).
        :param series_mean: The mean of the series.
        :param error_mean: The mean of the normal error of each past random shock.
        :param error_std: The standard deviation of the normal error of each past random shock.
        :param coefs_low: If specified, defines the lower bound of the coeficients to be generated. If left None,
                          then the coefs parameter must be specified.
        :param coefs_high: If specified, defines the upper bound of the coeficients to be generated. If left None,
                          then the coefs parameter must be specified.
        :param coefs: A list or dict with numeric keys of the coeficients to be employed. Must have size p. The i-th
                      element (if list) or key (if dict) correspond to the i-th coeficient.
                      If this is specified, coefs_low and coefs_high are ignored.
        :param error_event: An error event. If specified, it is used instead of error_mean and error_std.

        :return: The MA model.
        """

        # check coeficients
        if (coefs is None) and (coefs_low is None) and (coefs_high is None):
            raise ValueError("Either coefs or coefs_loe, coefs_high must be specified.")

        # error shocks
        if error_event is None:
            error_event = NormalEvent(error_mean, error_std)

        # Put the mean term first
        x = ConstantEvent(series_mean, parallel_events=[error_event])

        past = []

        # Add model terms
        for i in range(0, q):

            if coefs is not None:
                alpha = coefs[i]
            else:
                alpha = np.random.uniform(coefs_low, coefs_high)

            p = PastEvent(i + 1)
            past.append(p)
            x = x + p * ConstantEvent(alpha)

        # connect past events to the series to which they refer to
        for p in past:
            p.refers_to(error_event)

        return x

    @staticmethod
    @cli_model
    def arma(p, q, constant: float=0, error_mean: float=0, error_std: float=1,
             ar_coefs_low: float=-1, ar_coefs_high: float=1, ar_coefs: a_list=None,
             ma_coefs_low: float=-1, ma_coefs_high: float=1, ma_coefs: a_list=None):
        """
        Creates an ARMA model. This differs slightly from simply summing AR and MA models, because here a common
        normal error series is also provided

        :param p:
        :param q:
        :param constant:
        :param error_mean:
        :param error_std:
        :param ar_coefs_low:
        :param ar_coefs_high:
        :param ar_coefs:
        :param ma_coefs_low:
        :param ma_coefs_high:
        :param ma_coefs:
        :return:
        """

        # common error series
        error_event = NormalEvent(error_mean, error_std)

        m1 = ClassicModels.ar(p, constant=constant, coefs_low=ar_coefs_low, coefs_high=ar_coefs_high, coefs=ar_coefs,
                error_event=error_event)

        m2 = ClassicModels.ma(q, series_mean=0.0, coefs_low=ma_coefs_low, coefs_high=ma_coefs_high, coefs=ma_coefs,
                error_event=error_event)

        return m1 + m2


class BankingModels:

    @staticmethod
    @cli_model
    def salary_earner(salary: float=5000, payment_day: int=1, *, expense_mean: float=100.0, expense_sd: float=400.0):

        daily_normal_expense = NormalEvent(expense_mean, expense_sd)

        def aux(t, i, memory):

            actual_payment_day = shift_weekend_and_holidays(pd.Timestamp(t.year, t.month, payment_day),
                                                            direction='backward')

            if t.date() == actual_payment_day.date():
                memory['money'] = memory.get('money', 0.0) + salary
            else:
                memory['money'] = memory.get('money', 0.0) - daily_normal_expense.execute(t)

            return memory['money']

        model = LambdaEvent(aux, sub_events=[daily_normal_expense])
        return model


class EconomicModels:

    @staticmethod
    @cli_model
    def kondratiev_business_cycle(base: float = 0.0, growth_mean: float = 1, growth_sd: float = 2,
                             wave_period: float = 12, wave_amplitude: float= 0.05):
        """
        A naive interpretation of so-called "Kondratieve waves" business cycle theory.

        :param base: The initial economic condition.
        :param growth_mean: The mean of economic growth.
        :param growth_sd: The standard deviation of economic growth.
        :param wave_period: The period of the wave that modifies present conditions.
        :param wave_amplitude: The amplitude of the wave that modifies present conditions.
        :return:
        """
        we = WaveEvent(wave_period, wave_amplitude)
        return WalkEvent(base, NormalEvent(growth_mean, growth_sd)) * (ConstantEvent(1) + we)


class EcologyModels:

    @staticmethod
    @cli_model
    def predator_prey(n_predators=10, n_preys=40,
                      alpha=1.1, beta=0.02, delta=0.02, gamma=0.008):
        """
        Discrete version of the Lotka–Volterra equations for predator-prey model. The equations are:

          preys(t + 1)       = alpha*preys(t)           - beta*preys*predators(t)
          predators(t + 1)   = delta*preys*predators(t) - gamma*predators(t)

        :param n_predators: Initial number of predators.
        :param n_preys: Initial number of preys.
        :param alpha: Prey reproduction factor.
        :param beta: Effective killing factor for predators.
        :param delta: Effective multiplication factor for predators based on prey consumption.
        :param gamma: Natural death rate for predators.

        :return: A predators and a preys model.
        """

        # preys     = alpha*preys           - beta*preys*predators
        # predators = delta*preys*predators - gamma*predators

        past_preys = PastEvent(1, undefined_value=n_preys, name='Past Preys')
        past_predators = PastEvent(1, undefined_value=n_predators, name='Past Predators')

        preys = ClipEvent((ConstantEvent(alpha) * past_preys - ConstantEvent(beta)*past_preys*past_predators),
                          min_value=0.0,
                          name="Preys")

        predators = ClipEvent((ConstantEvent(delta) * past_preys * past_predators - ConstantEvent(gamma) * past_predators),
                              min_value=0.0,
                              name="Predators")

        preys.parallel_to(predators)
        predators.parallel_to(preys)

        past_preys.refers_to(preys)
        past_predators.refers_to(predators)

        return predators, preys



from setuptools import setup

setup(name='time-blender',
      version='0.1.2',
      description='A compositional time series generator.',
      url='https://github.com/paulosalem/time-blender',
      author='Paulo Salem',
      author_email='paulosalem@paulosalem.com',
      license='MIT',
      packages=['time_blender'],
      classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent"],
      include_package_data=True,
      install_requires=[
            'pandas', 'numpy', 'clize', 'pymc3', 'sigtools', 'matplotlib'
      ],
      scripts=['bin/time_blender'])


from time_blender.core import Event, RandomEvent
import time_blender.config as config
import pymc3 as pm
import random
import numpy as np

# Filters, connectors, etc.


# TODO DatetimeFilter  # weekends, specific hours, holidays, etc.

from time_blender.random_events import NormalEvent
from time_blender.util import PyMC3Utils


class Once(Event):
    """
    Executes and memoizes the result of an event, so that it is not recomputed later.
    """

    def __init__(self, event, name=None, parallel_events=None, push_down=False):
        super().__init__(name, parallel_events, push_down)
        self.event = event
        self.value = None

    def _execute(self, t, i):
        if self.value is None:
            self.value = self.event.execute(t)

        return self.value

    def reset(self):
        super().reset()
        self.value = None


class PastEvent(Event):
    def __init__(self, delay, undefined_value=0.0, refers_to=None, name=None, parallel_events=None, push_down=False):
        super().__init__(name, parallel_events, push_down)
        self.delay = delay
        self.undefined_value = undefined_value
        self.event = refers_to

    def refers_to(self, event):
        self.event = event
        return self

    def _execute(self, t, i):
        if self.event is None:
            raise ValueError("The event to which the present event refers to has not been defined yet.")
        else:
            pos = i - self.delay
            if pos >= 0:
                try:
                    v = self.event.value_at_pos(pos)

                except IndexError:
                    self.event.execute(t)
                    v = self.event.value_at_pos(pos)

                return v
            else:
                return self.undefined_value


class SeasonalEvent(Event):
    def __init__(self, event, base=0, year:int=None, month:int=None, day:int=None,
                 hour:int=None, minute:int=None, second:int=None,
                 name=None, parallel_events=None, push_down=False):

        self.event = event
        self.base = base
        self.year = year
        self.month = month
        self.day = day
        self.hour = hour
        self.minute = minute
        self.second = second

        super().__init__(name, parallel_events, push_down)

    def _execute(self, t, i):

        def aux_match(a, b, cont):

            if a is not None:
                if a == b:
                    return cont
                else:
                    return False
            else:
                return cont

        is_season = \
            aux_match(self.year, t.year,
                      aux_match(self.month, t.month,
                                aux_match(self.day, t.day,
                                          aux_match(self.hour, t.hour,
                                                    aux_match(self.minute, t.minute,
                                                              aux_match(self.second, t.second,
                                                                        True))))))
        if is_season:
            return self.event.execute(t)
        else:
            return self.base


class Choice(RandomEvent):

    def __init__(self, events, name=None, parallel_events=None, push_down=False):
        """

        :param events: When given a list of events, a uniform distribution will be assumed. When given a dict,
                       keys are events and values are their probabilities according to a categorical distribution.
        :param name:
        """

        if isinstance(events, dict):
            self.events = list(events.keys())
            self.probs = list(events.values())
            self.events_probs = events

        elif isinstance(events, list):
            self.events = events

            uni_prob = 1.0/len(events)
            self.events_probs = {e: uni_prob for e in events}
            self.probs = list(self.events_probs.values())

        name = self._default_name_if_none(name)

        # a function to specify the PyMC3 structure to use for choice
        def aux(model, obs):
            c = pm.Categorical(name, p=self.probs, observed=obs)
            all_pymc3_vars = [self._pymc3_model_variables_if_distribution(model, e) for e in events]
            return PyMC3Utils.multi_switch(all_pymc3_vars, c)

        super().__init__(aux, name, parallel_events, push_down)

    def sample_from_definition(self, t):
        raise NotImplementedError("Choice cannot be sampled directly from.")

    def sample_from_learned_distribution(self):
        raise NotImplementedError("Choice cannot be sampled directly from.")

    def choose(self, t):
        if self._should_use_learned_sample():
            idx = self.sample_from_learned_distribution()
        else:
            d = pm.Categorical.dist(p=self.probs)
            idx = d.random()

        return self.events[idx]

    def _execute(self, t, i):
        return self.choose(t).execute(t)


class Piecewise(Event):

    def __init__(self, events: list, t_separators:list=None,
                 name=None, parallel_events=None, push_down=False):

        self.events = events
        self.event_choice = Choice(events)
        self.t_separators = t_separators

        self._cur_separator_pos = 0
        self._cur_separator = None

        super().__init__(name, parallel_events, push_down)

    def _execute(self, t, i):
        if self._cur_separator_pos < len(self.t_separators):
            self._cur_separator = self.t_separators[self._cur_separator_pos]

            if i >= self._value_or_execute_if_event(self._cur_separator, t):
                self._cur_separator_pos += 1

        return self.events[self._cur_separator_pos].execute(t)


# TODO  class RandomPiecewise(RandomEvent):
#
# def __init__(self, events: list, t_separators:list=None, spacing_mean:float=None, spacing_sd:float=None, mode: str='fixed',
#                  name=None, parallel_events=None, random_seed=None):


class Top(RandomEvent):

    def __init__(self, event, resistance_probability, name=None, parallel_events=None, push_down=False):
        self.event = event
        self.resistance_probability = resistance_probability

        super().__init__(name, parallel_events, push_down)

# Trading: tops, bottoms, panics, shocks, news, etc.
# TODO

# Mean-reverting
# TODO
